# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT or BSD Licenses
# (See every file for its license)

"""
spyder.utils.help
=================

Configuration files for the Help plugin rich text mode
"""

import sys
from spyder.config.base import get_module_source_path
sys.path.insert(0, get_module_source_path(__name__))
